/* Nessus
 * Copyright (C) 1999, 2000 Renaud Deraison
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * In addition, as a special exception, Renaud Deraison
 * gives permission to link the code of this program with any
 * version of the OpenSSL library which is distributed under a
 * license identical to that listed in the included COPYING.OpenSSL
 * file, and distribute linked combinations including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * this file, you may extend this exception to your version of the
 * file, but you are not obligated to do so.  If you do not wish to
 * do so, delete this exception statement from your version.
 */

#include <includes.h>

#ifdef USE_GTK
#include <gtk/gtk.h>
#include "nessus_plugin.h"
#include "context.h"
#include "comm.h"
#include "prefs_dialog_plugins_prefs.h"
#include "listnotebook.h"
#include "readonly.h"
#include "error_dlg.h"
#include "ssh_keys_dialog.h"
#include "ssh_key_info_form.h"
#include "openvas_ssh_login.h"

#include "nessus_i18n.h"

static GtkWidget * pprefs_add_notebook_page(struct arglist *, char *, int);
static void pprefs_add_entry(struct arglist *, struct arglist *, char *, GtkWidget*);
static void pprefs_add_password(struct arglist *, struct arglist *, char *, GtkWidget*);
static void pprefs_add_file(struct arglist *, struct arglist *, char *, GtkWidget*);
static void pprefs_add_checkbox(struct arglist *, struct arglist *, char *, GtkWidget*);
static void pprefs_add_radio(struct arglist *, struct arglist *, char *, GtkWidget*);
static void pprefs_add_sshcredentials(struct arglist *, struct arglist *, char *, GtkWidget*, struct context* context);

/**
 * Adds and hooks up (or resets) the two notebook pages for Plugin Preferences 
 * and Credentials.
 * These pages are itself notebooks with dynamical content.
 * @param ctrls Arglist to hook the notebooks into.
 */
static void
create_plugin_prefs_containers(ctrls)
  struct arglist *ctrls;
{
  GtkWidget *listnotebook, *cred_listnotebook;

  listnotebook = listnotebook_new(FALSE, TRUE);
  read_only_set_recurse(listnotebook);
  gtk_container_set_border_width(GTK_CONTAINER(listnotebook), 10);

  cred_listnotebook = listnotebook_new(FALSE, TRUE);
  read_only_set_recurse(cred_listnotebook);
  gtk_container_set_border_width(GTK_CONTAINER(cred_listnotebook), 10);
  
  /* On the first call we have to add the values, on subsequent calls we
   * just set them */
  if (arg_get_value(ctrls, "PLUGIN_PREFS") != NULL)
  {
    arg_set_value(ctrls, "PLUGIN_PREFS", -1, listnotebook);
    arg_set_value(ctrls, "PLUGIN_CREDENTIALS", -1, cred_listnotebook);
  }
  else
  {
    arg_add_value(ctrls, "PLUGIN_PREFS", ARG_PTR, -1, listnotebook);
    arg_add_value(ctrls, "PLUGIN_CREDENTIALS", ARG_PTR, -1, cred_listnotebook);
  }
}

struct arglist *
prefs_dialog_plugins_prefs()
{
  struct arglist *ctrls = emalloc(sizeof(struct arglist));
  GtkWidget *frame, * cred_frame;
  GtkWidget *plugin_prefs, *plugin_credentials;

  frame = gtk_frame_new(_("Advanced Plugins preferences"));
  gtk_container_border_width(GTK_CONTAINER(frame), 10);
  gtk_widget_show(frame);
  arg_add_value(ctrls, "FRAME", ARG_PTR, -1, frame);
  read_only_set_recurse(frame);


  cred_frame = gtk_frame_new(_("Credentials"));
  gtk_container_border_width(GTK_CONTAINER(cred_frame), 10);
  gtk_widget_show(cred_frame);
  arg_add_value(ctrls, "FRAME_CREDENTIALS", ARG_PTR, -1, cred_frame);
  read_only_set_recurse(cred_frame);

  create_plugin_prefs_containers(ctrls);

  plugin_prefs = arg_get_value(ctrls, "PLUGIN_PREFS");
  gtk_container_add(GTK_CONTAINER(frame), plugin_prefs);
  gtk_widget_show(plugin_prefs);
  plugin_credentials = arg_get_value(ctrls, "PLUGIN_CREDENTIALS");
  gtk_container_add(GTK_CONTAINER(cred_frame), plugin_credentials);
  gtk_widget_show(plugin_credentials);

  return (ctrls);
}


void
prefs_dialog_plugins_prefs_read_only(struct arglist *ctrls, gboolean readonly)
{
  read_only_set_read_only(GTK_WIDGET(arg_get_value(ctrls, "FRAME")),
      readonly);
  read_only_set_read_only(GTK_WIDGET(arg_get_value(ctrls,
	      "FRAME_CREDENTIALS")), readonly);
}

/**
 * Fill the plugin preferences (and credentials) pages with the corresponding 
 * widgets and values.
 * Widgets are added dynamically, based on the plugin preference type, e.g.
 * for a PREF_CHECKBOX pprefs_add_checkbox() is called.
 * @param context The context to use.
 * @param ctrls Arglist holding the notebook page widgets.
 * @param plugins Plugin list.
 */
void
prefs_dialog_plugins_prefs_fill(context, ctrls, plugins)
  struct context *context;
  struct arglist *ctrls;
  struct nessus_plugin *plugins;
{
  struct arglist *pprefs = arg_get_value(context->prefs, "PLUGINS_PREFS");
  struct nessus_plugin *plugs = plugins;
  GtkWidget *notebook_vbox = NULL;

  while(plugs != NULL )
  {
    struct arglist *prefs;

    if((prefs = plugs->plugin_prefs) != NULL )
    {
      if (strcmp(plugs->family, "Credentials"))
      {
	notebook_vbox = pprefs_add_notebook_page(ctrls, plugs->name, 0);
      }
      else
      {
        notebook_vbox = pprefs_add_notebook_page(ctrls, plugs->name, 1);
      }
      // For each preference of this plugin, add the widget.
      while(prefs && prefs->next)
      {
	char *type, *value;
	GtkWidget *vbox;

        vbox = notebook_vbox;

	type = arg_get_value(prefs->value, "type");
	value = arg_get_value(prefs->value, "value");
	if(type)
	{
	  if(!strcmp(type, PREF_ENTRY))
	    pprefs_add_entry(pprefs, prefs, value, vbox);
	  else if(!strcmp(type, PREF_PASSWORD))
	    pprefs_add_password(pprefs, prefs, value, vbox);
	  else if(!strcmp(type, PREF_RADIO))
	    pprefs_add_radio(pprefs, prefs, value, vbox);
	  else if(!strcmp(type, PREF_CHECKBOX))
	    pprefs_add_checkbox(pprefs, prefs, value, vbox);
	  else if(!strcmp(type, PREF_FILE))
	    pprefs_add_file(pprefs, prefs, value, vbox);
          else if(!strcmp(type, PREF_SSH_CREDENTIALS))
            pprefs_add_sshcredentials(pprefs, prefs, value, vbox, context);
          else
            show_warning(_("%s asked for unknown preference type %s."), 
                        plugs->name, type);
	}
	prefs = prefs->next;
      }
    }
    plugs = plugs->next;
  }
}

/**
 * Clean up the plugin preferences and plugin
 * preferences widgets
 */
void
prefs_plugins_reset(ctrls, plugins)
  struct arglist *ctrls;
  struct nessus_plugin *plugins;
{
  struct arglist *prefs;

  while(plugins != NULL )
  {
    prefs =  plugins->plugin_prefs;
    if(prefs != NULL )
      while(prefs && prefs->next)
      {
	struct arglist *v;

	/*
	 * We keep the first three fields and destroy the others
	 */
	v = prefs->value;	/* value */
	v = v->next;		/* type */
	v = v->next;		/* fullname */
	arg_free(v->next);
	v->next = emalloc(sizeof(struct arglist));
	prefs = prefs->next;
      }
    plugins = plugins->next;
  }
}

/**
 * Redraw the plugins preferences
 */
void
prefs_plugins_prefs_redraw(context, ctrls)
  struct context *context;
  struct arglist *ctrls;
{
  GtkWidget *frame, *cred_frame;
  GtkWidget *plugin_prefs, *plugin_credentials;

  frame = arg_get_value(ctrls, "FRAME");
  plugin_prefs = arg_get_value(ctrls, "PLUGIN_PREFS");
  cred_frame = arg_get_value(ctrls, "FRAME_CREDENTIALS");
  plugin_credentials = arg_get_value(ctrls, "PLUGIN_CREDENTIALS");

  gtk_widget_hide(plugin_prefs);
  gtk_widget_hide(plugin_credentials);
  gtk_container_remove(GTK_CONTAINER(frame), plugin_prefs);
  gtk_container_remove(GTK_CONTAINER(cred_frame), plugin_credentials);

  plugin_prefs = context->plugin_prefs_widget;
  plugin_credentials = context->plugin_prefs_cred_widget;
  if(plugin_prefs != NULL && plugin_credentials != NULL)
  {
    arg_set_value(ctrls, "PLUGIN_PREFS", -1, plugin_prefs);
    arg_set_value(ctrls, "PLUGIN_CREDENTIALS", -1, plugin_credentials);
  }
  else
  {
    create_plugin_prefs_containers(ctrls);
    plugin_prefs = arg_get_value(ctrls, "PLUGIN_PREFS");
    plugin_credentials = arg_get_value(ctrls, "PLUGIN_CREDENTIALS");

    if((context->plugins && context->plugins->next)
	|| (context->scanners && context->scanners->next))
    {
      g_object_ref(plugin_prefs);
      g_object_ref(plugin_credentials);
      context->plugin_prefs_widget = plugin_prefs;
      context->plugin_prefs_cred_widget = plugin_credentials;

      prefs_plugins_reset(ctrls, context->scanners);
      prefs_plugins_reset(ctrls, context->plugins);
      prefs_dialog_plugins_prefs_fill(context, ctrls, context->scanners);
      prefs_dialog_plugins_prefs_fill(context, ctrls, context->plugins);
    }
  }
  gtk_container_add(GTK_CONTAINER(frame), plugin_prefs);
  /* It can happen that the GTK_PARENT_SENSITIVE flag is not set for the
   * children of plugin_prefs even though all parents are sensitive.
   * The effect of that is that the plugin prefs are insensitive even
   * though they shouldn't be.  Changing the sensitivity of plugin_prefs
   * twice fixes that.
   * It's not clear exactly why GTK_PARENT_SENSITIVE might be unset bit
   * is probably has to do with the gtk_container_add and
   * gtk_container_remove that done in this function and the fact that
   * the sensitivity of the parent widgets of plugin_prefs changes over
   * time.  This might be a GTK bug */
  gtk_widget_set_sensitive(plugin_prefs, FALSE);
  gtk_widget_set_sensitive(plugin_prefs, TRUE);
  gtk_widget_show(plugin_prefs);

  gtk_container_add(GTK_CONTAINER(cred_frame), plugin_credentials);
  /* cred_frame's children may also have an unset GTK_PARENT_SENSITIVE
   * flag so work around it in the same manner as above for plugin_prefs */
  gtk_widget_set_sensitive(cred_frame, FALSE);
  gtk_widget_set_sensitive(cred_frame, TRUE);
  gtk_widget_show(plugin_credentials);
}


/**
 * Adds a notebook page to either the plugin preference notebook or the 
 * credentials notebook and returns the vbox for it.
 * @param ctrls Arglist in which the other Gtk widgets are held.
 * @param name Name for the new notebook page.
 * @param credentials If 0 page is added to preference notebook, otherwise to 
 *                    credentials notebook.
 * @return The Gtk vbox nested in a Gtk scrolled window of the new page.
 */
static GtkWidget *
pprefs_add_notebook_page(ctrls, name, credentials)
  struct arglist *ctrls;
  char *name;
  int credentials;
{
  GtkWidget *vbox = NULL;
  GtkWidget *notebook = NULL;
  
  if (credentials == 0)
    notebook = arg_get_value(ctrls, "PLUGIN_PREFS");
  else
    notebook = arg_get_value(ctrls, "PLUGIN_CREDENTIALS");
  
  GtkWidget *s_window = gtk_scrolled_window_new(NULL, NULL);
  read_only_set_recurse(s_window);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(s_window),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_widget_show(s_window);
  vbox = gtk_vbox_new(FALSE, FALSE);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(s_window), vbox);
  gtk_widget_show(vbox);

  listnotebook_add_page(notebook, s_window, name, NULL);

  return vbox;
}

/**
 * Close the file dialog filew.
 * Callback for select_file.
 * @param filew GtkWidget to hide and destroy.
 * @param nul Ignored.
 * @return Always 0.
 */
static int
file_dialog_hide(GtkWidget * filew, GtkWidget * nul)
{
  gtk_widget_hide(filew);
  gtk_widget_destroy(filew);
  return 0;
}

/**
 * Callback for click on OK button in file selection dialog.
 * @see select_file
 * @param nul Ignored (callback).
 * @param filew Gtk File Selection widget.
 * @return Always 0.
 */
static int
file_selected(GtkWidget * nul, GtkWidget * filew)
{
  struct arglist *c;
  GtkWidget *entry;
  char *fname =
      (char *)gtk_file_selection_get_filename(GTK_FILE_SELECTION(filew));
  gtk_widget_hide(filew);
  c = gtk_object_get_data(GTK_OBJECT(filew), "data");
  entry = arg_get_value(c, "ENTRY");
  gtk_entry_set_text(GTK_ENTRY(entry), fname);
  return 0;
}

/**
 * Displays file selection dialog.
 * Callback for OK button is file_selected.
 * @see file_selected
 * @param b Ignored.
 * @param ctrl Arglist on which to operate (in callback).
 * @return Always 0.
 */
static int
select_file(GtkWidget * b, struct arglist *ctrls)
{

  GtkWidget *filew = gtk_file_selection_new(_("Select file"));

  g_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filew)->ok_button),
      "clicked", (GtkSignalFunc) file_selected, filew);
  g_signal_connect_swapped(GTK_OBJECT(GTK_FILE_SELECTION(filew)->
	  cancel_button), "clicked", GTK_SIGNAL_FUNC(file_dialog_hide),
      GTK_OBJECT(filew));
  gtk_object_set_data(GTK_OBJECT(filew), "data", ctrls);
  gtk_widget_show(filew);
  return 0;
}

/**
 * Returns "yes" or "no" if preference is of type ARG_INT.
 * @return ("yes" or "no") when pref of type int, value otherwise.
 */
static char* get_pref_value(struct arglist* prefname, struct arglist* prefvalue,
                            char* value)
{
 char* fullname = arg_get_value(prefname->value, "fullname");
 
 if(prefvalue)
 {
   int type;

   if((type = arg_get_type(prefvalue, fullname)) >= 0)
   {
     value = arg_get_value(prefvalue, fullname);
     if(type == ARG_INT)
     {
       if(value)
         value = strdup("yes");
       else
         value = strdup("no");
     }
   }
 }
 
 return value;
}


/**
 * Adds an entry to all the Comboboxes showing SSH keys.
 * To do this, all "next" and child contextes of ctx are recursively searched 
 * for plugins that have a ssh_credentials preferences, their combobox is 
 * unpacked and the new entry appended.
 * @param ctx Context to start looking for plugins with ssh_credentials preference.
 * @param newentry Text to add (should be key of ctx's sshkeys hash table).
 * @see context.sshkeys
 */
void
prefs_dialog_plugins_prefs_update_ssh_comboboxes(struct context* ctx,
                                                 char* newentry)
{
  if(ctx == NULL || newentry == NULL)
    return;
  
  struct nessus_plugin* plugins;
  struct arglist* pref;
  char* type;
  
  // Look in all Scopes + Global
  while(ctx && ctx->type < CONTEXT_REPORT)
  {
    prefs_dialog_plugins_prefs_update_ssh_comboboxes(ctx->children, newentry);
    if(ctx->type == CONTEXT_SCOPE || ctx->type == CONTEXT_GLOBAL)
    {
      // Look in all plugins for a SSH_LOGIN preference
      plugins = ctx->plugins;
      while(plugins != NULL)
      {
        pref = plugins->plugin_prefs;
        while(pref && pref->next)
        {
          type = (char *)arg_get_value(pref->value, "type");
          if(type != NULL && !strcmp(type, PREF_SSH_CREDENTIALS))
          {
            GtkWidget* combobox = arg_get_value(pref->value, "COMBOBOX");
            gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), newentry);
          }
          pref = pref->next;
        }
        plugins = plugins->next;
      }
    }
    ctx = ctx->next;
  }

  return;
}


/**
 * Helper function to set the active text of a combobox that has been 
 * initialized with gtk_combobox_new_text().
 * If value is not in the list of combobox entries, the first entry will be
 * set to be the active one.
 * @param combobox GtkComboBox to manipulate.
 * @param value String of the active-item-to-be.
 */
static void
text_combobox_set_active_text(GtkWidget* combobox, char* value)
{
  GtkTreeIter iter;
  gboolean valid;
  int idx = -1;
  GtkTreeModel *  treemdl = gtk_combo_box_get_model(GTK_COMBO_BOX(combobox));
  valid = gtk_tree_model_get_iter_first(treemdl, &iter);
  // Find the right item.
  while(valid)
  {
    char* entry;
    idx++;
    gtk_tree_model_get(treemdl, &iter, 0, &entry, -1);
    valid = gtk_tree_model_iter_next (treemdl, &iter);
    // If text is the text we want to be active, make it active and return.
    if(strcmp(entry, value) == 0)
    {
      break;
    }
    // If that was the last item, we apparently havent found any match
    if(valid == FALSE)
     idx = 0;
  }
  
  gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), idx);
  
}

/**
 * Fills the ssh_key_info_form form with values from an openvas_ssh_login.
 * The openvas_ssh_login is referenced by its name, which is the selected item
 * of a combobox. The information is then gathered ftom the Global contexts 
 * sshkeys Hashtable. To be used as a "change" signal callback of a combobox.
 * @param combobox Combobox to read active item from.
 * @param form The ssh_key_info_form to update.
 */
static void
show_sshinfo_cb(GtkWidget* combobox, ssh_key_info_form* form)
{
  openvas_ssh_login* sellogin = g_hash_table_lookup(Global->sshkeys, gtk_combo_box_get_active_text(GTK_COMBO_BOX(combobox)));
  ssh_key_info_form_fill(form, sellogin);
}

/**
 * Callback to add a key from a GHashTable to a Combobox.
 * @param key Will be added to combobox.
 * @param value --ignored-- (callback)
 */
static void
add_key_to_combobox_cb(char* key, gpointer value, GtkWidget* combobox)
{
  gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), key);
}
    

/**
 * Function to add a "ssh-credentials combo box" to select a ssh key.
 */
static void
pprefs_add_sshcredentials(struct arglist* pprefs, struct arglist* pref, 
                          char* value, GtkWidget* vbox, struct context* context)
{
  GtkWidget* combobox;
  GtkWidget* text;
  GtkWidget* hbox;
  GtkWidget* sshinfoframe;
  GtkWidget* managerbutton;
  char *name = pref->name;
  
  value = get_pref_value(pref, pprefs, value);
  
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 5);
  gtk_widget_show(hbox);

  text = gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 5);
  gtk_widget_show(text);

  combobox = gtk_combo_box_new_text ();
  gtk_combo_box_append_text (GTK_COMBO_BOX(combobox), NO_SSH_LOGIN_SELECTED);
  gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0);
  
  // First add frame, so that signal function is already set and set_active_text
  // will trigger filling the info frame.
  sshinfoframe = gtk_frame_new(_("SSH Key Information"));
  gtk_widget_show(sshinfoframe);
  gtk_box_pack_start_defaults(GTK_BOX(vbox), sshinfoframe);
  
  ssh_key_info_form* loginform = ssh_key_info_form_create();
  gtk_container_add (GTK_CONTAINER (sshinfoframe), loginform->box);
  
  g_signal_connect(GTK_OBJECT(combobox), "changed", 
                   (GtkSignalFunc) show_sshinfo_cb, loginform);
  
  // Add all the known accounts to combobox
  if(Global->sshkeys != NULL)
  {
    // TODO: If entries should be added sorted, we can generate a list of keys
    // (either update gtk or generate by hand) or consistently use a treemodel
    g_hash_table_foreach(Global->sshkeys, (GHFunc) add_key_to_combobox_cb, combobox);
    text_combobox_set_active_text(combobox, value);
  }
  
  gtk_box_pack_end(GTK_BOX(hbox), combobox, TRUE, TRUE, 5);
  
  //  Add and hook up button to open key management
  managerbutton = gtk_button_new_with_label(_("OpenVAS SSH Key Manager"));
  gtk_widget_show(managerbutton);
  gtk_box_pack_start_defaults(GTK_BOX(vbox), managerbutton);
  
  g_signal_connect(GTK_OBJECT(managerbutton), "clicked",
                   (GtkSignalFunc) ssh_manager_button_cb, 
                   (void*) NULL);
  
  arg_add_value(pref->value, "COMBOBOX", ARG_PTR, -1, combobox);
  
  gtk_widget_show(combobox);
  gtk_widget_show_all(sshinfoframe);
}


/**
 * Add "entry" widget corresponding to the "entry" plugin preference.
 */
static void
pprefs_add_entry(pprefs, pref, value, vbox)
  struct arglist *pprefs;
  struct arglist *pref;
  char *value;
  GtkWidget *vbox;
{
  GtkWidget *entry, *text, *box;
  char *name = pref->name;
  
  value = get_pref_value(pref, pprefs, value);
  
  box = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 5);
  gtk_widget_show(box);

  text = gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX(box), text, TRUE, TRUE, 5);
  gtk_widget_show(text);

  entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry), value);
  gtk_box_pack_end(GTK_BOX(box), entry, TRUE, TRUE, 5);
  gtk_widget_show(entry);
  arg_add_value(pref->value, "ENTRY", ARG_PTR, -1, entry);
}


/**
 * Add "password" widget corresponding to the "password" plugin preference.
 */
static void
pprefs_add_password(pprefs, pref, value, vbox)
  struct arglist *pprefs;
  struct arglist *pref;
  char *value;
  GtkWidget *vbox;
{
  GtkWidget *entry, *text, *box;
  char *name = pref->name;
  
  value = get_pref_value(pref, pprefs, value);
  
  box = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 5);
  gtk_widget_show(box);

  text = gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX(box), text, TRUE, TRUE, 5);
  gtk_widget_show(text);

  entry = gtk_entry_new();
  gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
  gtk_entry_set_text(GTK_ENTRY(entry), value);
  gtk_box_pack_end(GTK_BOX(box), entry, TRUE, TRUE, 5);
  gtk_widget_show(entry);
  arg_add_value(pref->value, "ENTRY", ARG_PTR, -1, entry);
}


/**
 * Add "file" widgets corresponding to the "file" plugin preference.
 */
static void
pprefs_add_file(pprefs, pref, value, vbox)
  struct arglist *pprefs;
  struct arglist *pref;
  char *value;
  GtkWidget *vbox;
{
  GtkWidget *entry, *text, *box;
  GtkWidget *hbox, *button;
  char *name = pref->name;
  
  value = get_pref_value(pref, pprefs, value);
  
  box = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 5);
  gtk_widget_show(box);

  text = gtk_label_new(name);
  gtk_box_pack_start(GTK_BOX(box), text, TRUE, TRUE, 5);
  gtk_widget_show(text);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_end(GTK_BOX(box), hbox, TRUE, TRUE, 5);
  gtk_widget_show(hbox);

  entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry), value);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 5);
  gtk_widget_show(entry);
  arg_add_value(pref->value, "ENTRY", ARG_PTR, -1, entry);

  button = gtk_button_new_with_label(_("Select..."));
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 5);
  g_signal_connect(GTK_OBJECT(button), "clicked",
      (GtkSignalFunc) select_file, pref->value);

  gtk_widget_show(button);
}


/**
 * Add radio buttons corresponding to the "radio" plugin preference.
 */
static void
pprefs_add_radio(pprefs, pref, value, vbox)
  struct arglist *pprefs;
  struct arglist *pref;
  char *value;
  GtkWidget *vbox;
{
  GtkWidget *orig;
  GtkWidget *button, *first_button;
  GtkWidget *label;
  char *t;
  GSList *list = NULL;
  char* def = NULL;
  
  def = get_pref_value(pref, pprefs, def);

  label = gtk_label_new(pref->name);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 5);
  gtk_widget_show(label);

  t = strchr(value, ';');
  if(t)
    t[0] = '\0';
  first_button = orig = gtk_radio_button_new_with_label(NULL, value);
  gtk_box_pack_start(GTK_BOX(vbox), orig, FALSE, FALSE, 5);
  gtk_object_set_data(GTK_OBJECT(orig), "name", value);
  gtk_widget_show(orig);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(orig), TRUE);
  value = t + sizeof(char);
  if(t)
    while(value)
    {
      if((t = strchr(value, ';')))
	t[0] = '\0';
      button =
	  gtk_radio_button_new_with_label(gtk_radio_button_group
	  (GTK_RADIO_BUTTON(orig)), value);
      gtk_object_set_data(GTK_OBJECT(button), "name", value);
      if(def && !strcmp(def, value))
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
      gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 5);
      gtk_widget_show(button);
      if(t)
	value = t + sizeof(char);
      else
	value = NULL;
    }
  list = gtk_radio_button_group(GTK_RADIO_BUTTON(orig));

  arg_add_value(pref->value, "RADIOBUTTONS", ARG_PTR, -1, list);
}


/**
 * Add a "checkbox" widget corresponding to the "checkbox" plugin preference.
 */
static void
pprefs_add_checkbox(pprefs, pref, value, vbox)
  struct arglist *pprefs;
  struct arglist *pref;
  char *value;
  GtkWidget *vbox;
{
  GtkWidget *box;
  GtkWidget *button;
  char *name = pref->name;
  char *def = NULL;

  def = get_pref_value(pref, pprefs, def);
      
  box = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 5);
  gtk_widget_show(box);

  button = gtk_check_button_new_with_label(estrdup(name));
  gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 5);
  gtk_widget_show(button);
  if(def)
  {
    if(!strcmp(def, "yes"))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
    else
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), FALSE);
  }
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), !strcmp(value, "yes"));
  arg_add_value(pref->value, "CHECKBOX", ARG_PTR, -1, button);
}
#endif
